"""Classes for monitoring the monitoring server stack.

    Covering mon-persister, mon-api and mon-thresh.
    Kafka, mysql, vertica and influxdb are covered by other detection plugins. Mon-notification uses statsd.
"""

import logging
import yaml

import monasca_setup.agent_config
import monasca_setup.detection
from monasca_setup.detection import find_process_cmdline, watch_process

log = logging.getLogger(__name__)


class MonAPI(monasca_setup.detection.Plugin):
    """Detect mon_api and setup monitoring."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        monasca_api = monasca_setup.detection.find_process_cmdline('monasca-api')
        if monasca_api is not None:
            # monasca-api can show up in urls and be an arg to this setup program, check port also
            # Find the right port from the config, this is specific to the Java version
            try:
                with open('/etc/monasca/api-config.yml', 'r') as config:
                    self.api_config = yaml.load(config.read())
            except Exception:
                log.exception('Failed parsing /etc/monasca/api-config.yml')
                self.available = False
                return
            api_port = self.api_config['server']['applicationConnectors'][0]['port']
            for conn in monasca_api.connections('inet'):
                if conn.laddr[1] == api_port:
                    self.available = True
                    return

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca api healthcheck")
        admin_port = self.api_config['server']['adminConnectors'][0]['port']
        return dropwizard_health_check('monitoring', 'api', 'http://localhost:{0}/healthcheck'.format(admin_port))

        # todo
        # log.info("\tEnabling the mon api metric collection")
        # http://localhost:8081/metrics

    def dependencies_installed(self):
        return True


class MonNotification(monasca_setup.detection.Plugin):
    """Detect the Monsaca notification engine and setup some simple checks."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        if find_process_cmdline('monasca-notification') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca Notification healthcheck")
        return watch_process(['monasca-notification'], 'monitoring', 'notification', exact_match=False)

    def dependencies_installed(self):
        return True


class MonPersister(monasca_setup.detection.Plugin):
    """Detect mon_persister and setup monitoring."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        if find_process_cmdline('monasca-persister') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca persister healthcheck")
        return dropwizard_health_check('monitoring', 'persister', 'http://localhost:8091/healthcheck')

        # todo
        # log.info("\tEnabling the mon persister metric collection")
        # http://localhost:8091/metrics

    def dependencies_installed(self):
        return True


class MonThresh(monasca_setup.detection.Plugin):
    """Detect the running mon-thresh and monitor."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        if find_process_cmdline('backtype.storm.daemon') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tWatching the mon-thresh process.")
        config = monasca_setup.agent_config.Plugins()
        for process in ['backtype.storm.daemon.nimbus', 'backtype.storm.daemon.supervisor', 'backtype.storm.daemon.worker']:
            if find_process_cmdline(process) is not None:
                config.merge(watch_process([process], 'monitoring', 'storm', exact_match=False))
        return config

    def dependencies_installed(self):
        return True


def dropwizard_health_check(service, component, url):
    """Setup a dropwizard heathcheck to be watched by the http_check plugin."""
    config = monasca_setup.agent_config.Plugins()
    config['http_check'] = {'init_config': None,
                            'instances': [{'name': "{0}-{1} healthcheck".format(service, component),
                                           'url': url,
                                           'timeout': 1,
                                           'include_content': False,
                                           'dimensions': {'service': service, 'component': component}}]}
    return config



""" Metric data types
"""
from collections import namedtuple
import logging
from time import time

from monasca_agent.common.exceptions import Infinity, UnknownValue


log = logging.getLogger(__name__)


class Measurement(object):
    def __init__(self, name, timestamp, value, dimensions, delegated_tenant=None, value_meta=None):
        self.name = name
        self.timestamp = timestamp
        self.value = value
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None


class MetricTypes(object):
    GAUGE = 'gauge'
    COUNTER = 'counter'
    RATE = 'rate'


class Metric(object):

    """
    A base metric class that accepts points, slices them into time intervals
    and performs roll-ups within those intervals.
    """

    def sample(self, value, sample_rate, timestamp=None):
        """ Add a point to the given metric. """
        raise NotImplementedError()

    def flush(self, timestamp):
        """ Flush all metrics up to the given timestamp. """
        raise NotImplementedError()


class Gauge(Metric):

    """ A metric that tracks a value at particular points in time. """

    def __init__(self, formatter, name, dimensions,
                 hostname, device_name, delegated_tenant=None,
                 value_meta=None):
        self.formatter = formatter
        self.name = name
        self.value = None
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        self.hostname = hostname
        self.device_name = device_name
        self.timestamp = time()
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None

    def sample(self, value, sample_rate, timestamp=None):
        self.value = value
        self.timestamp = timestamp

    def flush(self, timestamp):
        if self.value is not None:
            value = self.value
            self.value = None
            return [self.formatter(metric=self.name,
                                   timestamp=self.timestamp or timestamp,
                                   value=value,
                                   dimensions=self.dimensions,
                                   delegated_tenant=self.delegated_tenant,
                                   hostname=self.hostname,
                                   device_name=self.device_name,
                                   metric_type=MetricTypes.GAUGE,
                                   value_meta=self.value_meta)]
        else:
            return []


class Counter(Metric):

    """ A metric that tracks a counter value. """

    def __init__(self, formatter, name, dimensions,
                 hostname, device_name, delegated_tenant=None,
                 value_meta=None):
        self.formatter = formatter
        self.name = name
        self.value = 0
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        self.hostname = hostname
        self.device_name = device_name
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None

    def sample(self, value, sample_rate, timestamp=None):
        try:
          self.value += value * int(1 / sample_rate)
        except TypeError:
          log.error("metric {} value {} sample_rate {}".format(self.name, value, sample_rate))

    def flush(self, timestamp):
        if self.value is not None:
            value = self.value
            self.value = 0
            return [self.formatter(metric=self.name,
                                   value=value,
                                   timestamp=timestamp,
                                   dimensions=self.dimensions,
                                   delegated_tenant=self.delegated_tenant,
                                   hostname=self.hostname,
                                   device_name=self.device_name,
                                   metric_type=MetricTypes.RATE,
                                   value_meta=self.value_meta)]
        else:
            return []


class Histogram(Metric):

    """ A metric to track the distribution of a set of values. """

    def __init__(self, formatter, name, dimensions,
                 hostname, device_name, delegated_tenant=None,
                 value_meta=None):
        self.formatter = formatter
        self.name = name
        self.count = 0
        self.samples = []
        self.percentiles = [0.95]
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        self.hostname = hostname
        self.device_name = device_name
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None

    def sample(self, value, sample_rate, timestamp=None):
        self.count += int(1 / sample_rate)
        self.samples.append(value)

    def flush(self, timestamp):
        metrics = []
        if not self.count:
            return metrics

        self.samples.sort()
        length = len(self.samples)

        max_ = self.samples[-1]
        med = self.samples[int(round(length / 2 - 1))]
        avg = sum(self.samples) / float(length)

        metric_aggrs = [
            ('max', max_, MetricTypes.GAUGE),
            ('median', med, MetricTypes.GAUGE),
            ('avg', avg, MetricTypes.GAUGE),
            ('count', self.count, MetricTypes.RATE)
        ]

        metrics.extend(self.formatter(hostname=self.hostname,
                                       device_name=self.device_name,
                                       dimensions=self.dimensions,
                                       delegated_tenant=self.delegated_tenant,
                                       metric='%s.%s' % (self.name, suffix),
                                       value=value,
                                       timestamp=timestamp,
                                       metric_type=metric_type,
                                       value_meta=self.value_meta
                                       ) for suffix, value, metric_type in metric_aggrs)

        for p in self.percentiles:
            val = self.samples[int(round(p * length - 1))]
            name = '%s.%spercentile' % (self.name, int(p * 100))
            metrics.append(self.formatter(hostname=self.hostname,
                                          dimensions=self.dimensions,
                                          delegated_tenant=self.delegated_tenant,
                                          metric=name,
                                          value=val,
                                          timestamp=timestamp,
                                          metric_type=MetricTypes.GAUGE,
                                          value_meta=self.value_meta))
        # Reset our state.
        self.samples = []
        self.count = 0

        return metrics


class Set(Metric):

    """ A metric to track the number of unique elements in a set. """

    def __init__(self, formatter, name, dimensions,
                 hostname, device_name, delegated_tenant=None,
                 value_meta=None):
        self.formatter = formatter
        self.name = name
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        self.hostname = hostname
        self.device_name = device_name
        self.values = set()
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None

    def sample(self, value, sample_rate, timestamp=None):
        self.values.add(value)

    def flush(self, timestamp):
        metrics = []
        if not self.values:
            return []
        else:
            values = self.values.copy()
            self.values = set()

        return [self.formatter(hostname=self.hostname,
                               device_name=self.device_name,
                               dimensions=self.dimensions,
                               delegated_tenant=self.delegated_tenant,
                               metric=self.name,
                               value=len(values),
                               timestamp=timestamp,
                               metric_type=MetricTypes.GAUGE,
                               value_meta=self.value_meta)]


class Rate(Metric):

    """ Track the rate of metrics over each flush interval """

    def __init__(self, formatter, name, dimensions,
                 hostname, device_name, delegated_tenant=None,
                 value_meta=None):
        self.formatter = formatter
        self.name = name
        self.dimensions = dimensions.copy()
        self.delegated_tenant = delegated_tenant
        self.hostname = hostname
        self.device_name = device_name
        self.samples = []
        if value_meta:
            self.value_meta = value_meta.copy()
        else:
            self.value_meta = None

    def sample(self, value, sample_rate, timestamp=None):
        if not timestamp:
            timestamp = time()
        self.samples.append((int(timestamp), value))

    def _rate(self, sample1, sample2):
        rate_interval = sample2[0] - sample1[0]
        if rate_interval == 0:
            log.warn('Metric %s has an interval of 0. Not flushing.' % self.name)
            raise Infinity()

        delta = sample2[1] - sample1[1]
        if delta < 0:
            log.info('Metric %s has a rate < 0. Counter may have been Reset.' % self.name)
            raise UnknownValue()

        return (delta / float(rate_interval))

    def flush(self, timestamp):
        if len(self.samples) < 2:
            return []

        val = self._rate(self.samples[-2], self.samples[-1])
        self.samples = self.samples[-1:]

        return [self.formatter(hostname=self.hostname,
                               device_name=self.device_name,
                               dimensions=self.dimensions,
                               delegated_tenant=self.delegated_tenant,
                               metric=self.name,
                               value=val,
                               timestamp=timestamp,
                               metric_type=MetricTypes.GAUGE,
                               value_meta=self.value_meta)]



import os

import monasca_setup.detection

DIAG_COMMAND = '/usr/local/bin/diagnostics'
CHECKER_COMMAND = '/usr/local/bin/swift-checker'

diag_attributes = ['check_mounts', 'disk_monitoring', 'file_ownership',
                   'network_interface', 'ping_hosts', 'drive_audit']

checker_attributes = ['diskusage', 'healthcheck', 'replication']


def get_instances(command, attributes):
    """
    Fetch instances per command type and attribute
    """
    instances = []
    for attribute in attributes:
        cmd = dict()
        cmd['name'] = '{0}.{1}'.format('swift', attribute)
        cmd['check_command'] = '{0} --{1}'.format(command, attribute)
        cmd['check_interval'] = 60
        cmd['dimensions'] = {'service': 'swift'}
        instances.append(cmd)
    return instances


def get_config_instances():
    """
    Fetch all instances per command type
    """
    config_instances = []
    config_instances.extend(get_instances(DIAG_COMMAND, diag_attributes))
    config_instances.extend(get_instances(CHECKER_COMMAND, checker_attributes))
    return config_instances


class Swift(monasca_setup.detection.ServicePlugin):

    """Detect Swift daemons and setup configuration to monitor them.

    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'object-storage',
            'process_names': ['swift-container-updater', 'swift-account-auditor',
                              'swift-object-replicator', 'swift-container-replicator',
                              'swift-object-auditor', 'swift-container-auditor',
                              'swift-account-reaper', 'swift-container-sync',
                              'swift-account-replicator', 'swift-object-updater',
                              'swift-object-server', 'swift-account-server',
                              'swift-container-server', 'swift-proxy-server'],
            'service_api_url': 'http://localhost:8080/healthcheck',
            'search_pattern': '.*OK.*'
        }

        super(Swift, self).__init__(service_params)

    def build_config(self):
        config = super(Swift, self).build_config()

        if os.path.exists(DIAG_COMMAND) and \
                os.path.exists(CHECKER_COMMAND):
            config['nagios_wrapper'] = {'init_config': None,
                                        'instances': get_config_instances(),
                                        }

        return config



import psutil
import logging

import monasca_agent.collector.checks as checks

log = logging.getLogger(__name__)


class Cpu(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Cpu, self).__init__(name, init_config, agent_config)

    def check(self, instance):
        """Capture cpu stats

        """
        dimensions = self._set_dimensions(None, instance)

        if instance is not None:
            send_rollup_stats =  instance.get("send_rollup_stats", False)
        else:
            send_rollup_stats =  False

        cpu_stats = psutil.cpu_times_percent(percpu=False)
        num_of_metrics = self._format_results(cpu_stats.user + cpu_stats.nice,
                                              cpu_stats.system + cpu_stats.irq + cpu_stats.softirq,
                                              cpu_stats.iowait,
                                              cpu_stats.idle,
                                              cpu_stats.steal,
                                              dimensions)
        if send_rollup_stats:
            self.gauge('cpu.total_logical_cores', psutil.cpu_count(logical=True), dimensions)
            num_of_metrics += 1
        log.debug('Collected {0} cpu metrics'.format(num_of_metrics))


    def _format_results(self, us, sy, wa, idle, st, dimensions):
        data = {'cpu.user_perc': us,
                'cpu.system_perc': sy,
                'cpu.wait_perc': wa,
                'cpu.idle_perc': idle,
                'cpu.stolen_perc': st}

        for key in data.keys():
            if data[key] is None:
                del data[key]

        [self.gauge(key, value, dimensions) for key, value in data.iteritems()]

        return len(data)



import logging

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class HostAlive(monasca_setup.detection.ArgsPlugin):
    """ Setup an host_alive check according to the passed in args.
        Despite being a detection plugin this plugin does no detection and will be a noop without arguments.
        Expects two space seperated arguments hostname and type. Type can be either 'ssh' or 'ping'. For example:
        'monasca-setup -d hostalive -a "hostname=remotebox type=ping"'
    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        self.available = self._check_required_args(['hostname', 'type'])

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        log.info("\tEnabling {type} host check for {hostname}".format(**self.args))
        # Since the naming in the args and in the config don't match build_instance is only good for dimensions
        instance = self._build_instance([])
        instance.update({'name': "{hostname} {type}".format(**self.args),
                    'host_name': self.args['hostname'],
                    'alive_test': self.args['type']})
        config['host_alive'] = {'init_config': None, 'instances': [instance]}

        return config




import logging
import os
import re

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class Ntp(monasca_setup.detection.Plugin):
    """Detect NTP daemon and setup configuration to monitor them.
    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        if monasca_setup.detection.find_process_cmdline('ntp') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        log.info("\tEnabling the ntp plugin")
        if os.path.exists('/etc/ntp.conf'):
            server = re.compile('server (.*)')
            with open('/etc/ntp.conf', 'r') as ntp_config:
                match = server.search(ntp_config.read())
            if match is None:
                ntp_server = 'pool.ntp.org'
            else:
                ntp_server = match.group(1)
        else:
            ntp_server = 'pool.ntp.org'
        config['ntp'] = {'init_config': None, 'instances': [{'name': ntp_server, 'host': ntp_server}]}

        return config

    def dependencies_installed(self):
        try:
            import ntplib
        except ImportError:
            return False
        else:
            return True



# Core modules
import logging
import socket
import system.win32 as w32
import threading
import time

import monasca_agent.common.check_status as check_status
import monasca_agent.common.metrics as metrics
import monasca_agent.common.util as util


log = logging.getLogger(__name__)


MAX_THREADS_COUNT = 50
MAX_COLLECTION_TIME = 30
MAX_CPU_PCT = 10
FLUSH_LOGGING_PERIOD = 10
FLUSH_LOGGING_INITIAL = 5


class Collector(util.Dimensions):

    """The collector is responsible for collecting data from each check and

    passing it along to the emitters, who send it to their final destination.
    """

    def __init__(self, agent_config, emitter, checksd=None):
        super(Collector, self).__init__(agent_config)
        self.agent_config = agent_config
        self.os = util.get_os()
        self.plugins = None
        self.emitter = emitter
        socket.setdefaulttimeout(15)
        self.run_count = 0
        self.continue_running = True
        self.initialized_checks_d = []
        self.init_failed_checks_d = []

        # add windows system checks
        if self.os == 'windows':
            self._checks = [w32.Disk(log),
                            w32.IO(log),
                            w32.Processes(log),
                            w32.Memory(log),
                            w32.Network(log),
                            w32.Cpu(log)]
        else:
            self._checks = []

        if checksd:
            # is of type {check_name: check}
            self.initialized_checks_d = checksd['initialized_checks']
            # is of type {check_name: {error, traceback}}
            self.init_failed_checks_d = checksd['init_failed_checks']

    def _emit(self, payload):
        """Send the payload via the emitter.
        """
        statuses = []
        # Don't try to send to an emitter if we're stopping/
        if self.continue_running:
            name = self.emitter.__name__
            emitter_status = check_status.EmitterStatus(name)
            try:
                self.emitter(payload, log, self.agent_config['forwarder_url'])
            except Exception as e:
                log.exception("Error running emitter: %s" % self.emitter.__name__)
                emitter_status = check_status.EmitterStatus(name, e)
            statuses.append(emitter_status)
        return statuses

    def _set_status(self, check_statuses, emitter_statuses, collect_duration):
        try:
            check_status.CollectorStatus(check_statuses, emitter_statuses).persist()
        except Exception:
            log.exception("Error persisting collector status")

        if self.run_count <= FLUSH_LOGGING_INITIAL or self.run_count % FLUSH_LOGGING_PERIOD == 0:
            log.info("Finished run #%s. Collection time: %ss." %
                     (self.run_count, round(collect_duration, 2)))
            if self.run_count == FLUSH_LOGGING_INITIAL:
                log.info("First flushes done, next flushes will be logged every %s flushes." %
                         FLUSH_LOGGING_PERIOD)

        else:
            log.debug("Finished run #%s. Collection time: %ss." %
                      (self.run_count, round(collect_duration, 2),))

    def collector_stats(self, num_metrics, num_events, collection_time):
        metrics = {}
        thread_count = threading.active_count()
        metrics['monasca.thread_count'] = thread_count
        if thread_count > MAX_THREADS_COUNT:
            log.warn("Collector thread count is high: %d" % thread_count)

        metrics['monasca.collection_time_sec'] = collection_time
        if collection_time > MAX_COLLECTION_TIME:
            log.info("Collection time (s) is high: %.1f, metrics count: %d, events count: %d" %
                     (collection_time, num_metrics, num_events))

        return metrics

    def run(self):
        """Collect data from each check and submit their data.

        There are currently two types of checks the system checks and the configured ones from checks_d
        """
        timer = util.Timer()
        self.run_count += 1
        log.debug("Starting collection run #%s" % self.run_count)

        metrics_list = []

        timestamp = time.time()
        events = {}

        if self.os == 'windows':  # Windows uses old style checks.
            for check_type in self._checks:
                try:
                    for name, value in check_type.check().iteritems():
                        metrics_list.append(metrics.Measurement(name,
                                                                timestamp,
                                                                value,
                                                                self._set_dimensions(None),
                                                                None))
                except Exception:
                    log.exception('Error running check.')
        else:
            for check_type in self._checks:
                metrics_list.extend(check_type.check())

        # checks_d checks
        checks_d_metrics, checks_d_events, checks_statuses = self.run_checks_d()
        metrics_list.extend(checks_d_metrics)
        events.update(checks_d_events)

        # Store the metrics and events in the payload.
        collect_duration = timer.step()

        dimensions = {'component': 'monasca-agent'}
        # Add in metrics on the collector run
        for name, value in self.collector_stats(len(metrics_list), len(events),
                                                collect_duration).iteritems():
            metrics_list.append(metrics.Measurement(name,
                                                    timestamp,
                                                    value,
                                                    self._set_dimensions(dimensions),
                                                    None))
        emitter_statuses = self._emit(metrics_list)

        # Persist the status of the collection run.
        self._set_status(checks_statuses, emitter_statuses, collect_duration)

    def run_checks_d(self):
        """Run defined checks_d checks.

        returns a list of Measurements, a dictionary of events and a list of check statuses.
        """
        measurements = []
        events = {}
        check_statuses = []
        for check in self.initialized_checks_d:
            if not self.continue_running:
                return
            log.info("Running check %s" % check.name)
            instance_statuses = []
            metric_count = 0
            event_count = 0
            try:
                # Run the check.
                instance_statuses = check.run()

                # Collect the metrics and events.
                current_check_metrics = check.get_metrics()
                current_check_events = check.get_events()

                # Save them for the payload.
                measurements.extend(current_check_metrics)
                if current_check_events:
                    if check.name not in events:
                        events[check.name] = current_check_events
                    else:
                        events[check.name] += current_check_events

                # Save the status of the check.
                metric_count = len(current_check_metrics)
                event_count = len(current_check_events)
            except Exception:
                log.exception("Error running check %s" % check.name)

            status_check = check_status.CheckStatus(check.name, instance_statuses, metric_count, event_count,
                                                    library_versions=check.get_library_info())
            check_statuses.append(status_check)

        for check_name, info in self.init_failed_checks_d.iteritems():
            if not self.continue_running:
                return
            status_check = check_status.CheckStatus(check_name, None, None, None,
                                                    init_failed_error=info['error'],
                                                    init_failed_traceback=info['traceback'])
            check_statuses.append(status_check)

        return measurements, events, check_statuses

    def stop(self):
        """Tell the collector to stop at the next logical point.
        """
        # This is called when the process is being killed, so
        # try to stop the collector as soon as possible.
        # Most importantly, don't try to submit to the emitters
        # because the forwarder is quite possibly already killed
        # in which case we'll get a misleading error in the logs.
        # Best to not even try.
        self.continue_running = False
        for check in self.initialized_checks_d:
            check.stop()



"""Classes for detection of running resources to be monitored.

    Detection classes should be platform independent
"""
import sys
import logging

log = logging.getLogger(__name__)


class Plugin(object):
    """Abstract class implemented by the monasca-agent plugin detection classes. """

    def __init__(self, template_dir, overwrite=True, args=None):
        self.available = False
        self.template_dir = template_dir
        self.dependencies = ()
        self.overwrite = overwrite
        self.args = None
        if args is not None:
            try:
                # Turn 'hostname=host type=ping' to dictionary {'hostname': 'host', 'type': 'ping'}
                self.args = dict([a.split('=') for a in args.split()])
            except Exception:
                log.exception('Error parsing detection arguments')
                sys.exit(1)
        self._detect()

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        raise NotImplementedError

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        raise NotImplementedError

    def dependencies_installed(self):
        """Return True if dependencies are installed.

        """
        raise NotImplementedError

    @property
    def name(self):
        """Return _name if set otherwise the class name.

        """
        if '_name' in self.__dict__:
            return self._name
        else:
            return self.__class__.__name__



""" Util functions to assist in detection.
"""
import glob
import imp
import inspect
import logging
import os
import psutil
import subprocess

from monasca_setup import agent_config
from plugin import Plugin
from subprocess import Popen, PIPE, CalledProcessError

log = logging.getLogger(__name__)


# check_output was introduced in python 2.7, function added
# to accommodate python 2.6
try:
    check_output = subprocess.check_output
except AttributeError:
    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be overridden.')
        process = Popen(stdout=PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise CalledProcessError(retcode, cmd)
        return output


def find_plugins(custom_path):
    """ Find and import all detection plugins. It will look in detection/plugins dir of the code as well as custom_path

    :param custom_path: An additional path to search for detection plugins
    :return: A list of imported detection plugin classes.
    """

    # This was adapted from what monasca_agent.common.util.load_check_directory
    plugin_paths = glob.glob(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'plugins', '*.py'))
    plugin_paths.extend(glob.glob(os.path.join(custom_path, '*.py')))

    plugins = []

    for plugin_path in plugin_paths:
        if os.path.basename(plugin_path) == '__init__.py':
            continue
        try:
            plugin = imp.load_source(os.path.splitext(os.path.basename(plugin_path))[0], plugin_path)
        except Exception:
            log.exception('Unable to import detection plugin {0}'.format(plugin_path))

        # Verify this is a subclass of Plugin
        classes = inspect.getmembers(plugin, inspect.isclass)
        for _, clsmember in classes:
            if Plugin == clsmember:
                continue
            if issubclass(clsmember, Plugin):
                plugins.append(clsmember)

    return plugins


def find_process_cmdline(search_string):
    """Simple function to search running process for one with cmdline containing.
    """
    for process in psutil.process_iter():
        for arg in process.cmdline():
            if arg.find(search_string) != -1:
                return process

    return None


def find_process_name(pname):
    """Simple function to search running process for one with pname.
    """
    for process in psutil.process_iter():
        if pname == process.name():
            return process

    return None


def find_addr_listening_on_port(port):
    """Return the IP address which is listening on the specified TCP port."""
    for conn in psutil.net_connections(kind='tcp'):
        if conn.laddr[1] == port and conn.status == psutil.CONN_LISTEN:
            return conn.laddr[0].lstrip("::ffff:")


def watch_process(search_strings, service=None, component=None, exact_match=True, detailed=False):
    """Takes a list of process search strings and returns a Plugins object with the config set.
        This was built as a helper as many plugins setup process watching
    """
    config = agent_config.Plugins()
    parameters = {'name': search_strings[0],
                  'detailed': detailed,
                  'exact_match': exact_match,
                  'search_string': search_strings}

    dimensions = _get_dimensions(service, component)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['process'] = {'init_config': None,
                         'instances': [parameters]}
    return config


def service_api_check(name, url, pattern, service=None, component=None):
    """Setup a service api to be watched by the http_check plugin.
    """
    config = agent_config.Plugins()
    parameters = {'name': name,
                  'url': url,
                  'match_pattern': pattern,
                  'timeout': 10,
                  'use_keystone': True}

    dimensions = _get_dimensions(service, component)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['http_check'] = {'init_config': None,
                            'instances': [parameters]}

    return config


def _get_dimensions(service, component):
    dimensions = {}
    # If service parameter is set in the plugin config, add the service dimension which
    # will override the service in the agent config
    if service:
        dimensions.update({'service': service})

    # If component parameter is set in the plugin config, add the component dimension which
    # will override the component in the agent config
    if component:
        dimensions.update({'component': component})

    return dimensions
